#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <time.h>

#define 	MAX_NAME 	64		// MAX Name Size Of A Poet
#define 	MAX_HASH 	137		// MAX Hash Table Size
#define 	MAX_POET 	133		// MAX Poet Entry to Table
#define 	MNO_POEM 	50		// MAX Poet Number That Are Not in Hash
#define 	MAX_CHAR 	20000	// MAX Poem Size

using namespace std;



/***********************************************
 DATA STRUCTURE : Node
 MISSION 		: Used to store names of the 
 				  poet and the keys to poems
***********************************************/
typedef struct hash_node{
	int 	key;
	bool 	used;
	char	name[MAX_NAME];
}Node;

/***********************************************
 DATA STRUCTURE : Hash Table
 MISSION 		: Used to cretae the Hash Table
***********************************************/
typedef struct hash_table{
	Node node[MAX_HASH];
}Table;




/***********************************************
 FUNCTION 	: toDecimal
 MISSION 	: Counts The Decimal Number
***********************************************/
int toDecimal(int n){ //  6 = toDecimal(123456);
	if(n == 0)
		return 0;
	return 1 + toDecimal(n/10);
}

/***********************************************
 FUNCTION 	: ReadData
 MISSION 	: Reads the Poets From Database
***********************************************/
bool ReadData(Table *tmp, Table *noPoem){
	cout << "---------------------------------------------------------" << endl;
	cout << " Reading All The Poets From Database..." << endl;
	cout << "---------------------------------------------------------" << endl;
	FILE *f;
	int length, total = 0;
	int i;
	char name[MAX_CHAR];
	if(f = fopen("database.txt","r")){
		// READ THE POETS AND THE ADDRESS OF THEIR POEMS
		for(int j=0; j<MAX_POET; j++){
			fscanf(f, "%d", &length);
			total += length + toDecimal(length);
			if(feof(f))
				break;
			for(i = 0; i<length; i++)
				fscanf(f, "%c", &name[i]);
			name[i] = '\0';
			(*tmp).node[j].key = total;
			strcpy((*tmp).node[j].name, name);
			
			fscanf(f, "%d", &length);
			total += length + toDecimal(length);
			for(i = 0; i<length; i++){
				fscanf(f, "%c", &name[i]);
				if(name[i] == '\n') i++;
			}
		}
		// READ THE POETS THAT HAVE NO POEMS IN HASH TABLE
		for(int j=0; j<MNO_POEM; j++){
			fscanf(f, "%d", &length);
			total += length + toDecimal(length);
			if(feof(f))
				break;
			for(i = 0; i<length; i++)
				fscanf(f, "%c", &name[i]);
			name[i] = '\0';
			strcpy((*noPoem).node[j].name, name);
		}
		fclose(f);
		return true;
	}else
		return false;
}

/***********************************************
 FUNCTION 	: PrintTmpData
 MISSION 	: Prints the Poets From 
 			  Temporary List
***********************************************/
void PrintTmpData(Table *tmp, Table *noPoem){
	cout << endl << "---------------------------------------------------------" << endl;
	for(int i=0; i<MAX_POET; i++)
		cout << i+1 << "\t:\t"  << (*tmp).node[i].name << endl;
	cout << endl << "---------------------------------------------------------" << endl;
}

/***********************************************
              INSERT FUNCTIONS                */
/***********************************************
 FUNCTION 	: InsertLinear
 MISSION 	: Inserts Poets To Linear 
 			  Hash Table
 FORMULA 	: { 193.NAME[i] + (i+1).227 } % 233
 COLLISION	: NODE --> NODE + 1
***********************************************/
int InsertLinear(Table *table, const char name[], const int key){
	int nod = 0, collision = 0;
	int strlength = strlen(name);
	for(int m = 0; m < strlength; m++)
		nod += ((193*abs((int)(name[m])) + (m+1)*227)%233)%MAX_HASH;
	nod %= MAX_HASH;
	while((*table).node[nod].used){
		nod = (nod + 1) % MAX_HASH;
		collision++;
	}
	(*table).node[nod].key = key;
	(*table).node[nod].used = true;
	strcpy((*table).node[nod].name, name);
	return collision + 1;
}

/***********************************************
 FUNCTION 	: InsertQuadratic
 MISSION 	: Inserts Poets To Quadratic 
 			  Hash Table
 FORMULA 	: { 181.NAME[i] + (i+1).233 } % 199
 COLLISION	: NODE --> NODE + (179.i + 211.i^2)
***********************************************/
int InsertQuadratic(Table *table, const char name[], const int key){
	int nod = 0, collision = 0, w;
	int strlength = strlen(name);
	for(int m = 0; m < strlength; m++)
		nod += ((181*abs((int)(name[m])) + (m+1)*233)%199)%MAX_HASH;
	nod %= MAX_HASH;
	w = 0;
	while((*table).node[nod].used){
		w++;
		nod = (nod + 199*w + 211*w*w) % MAX_HASH;
		collision++;
	}
	(*table).node[nod].key = key;
	(*table).node[nod].used = true;
	strcpy((*table).node[nod].name, name);
	return collision + 1;
}

/***********************************************
 FUNCTION 	: InsertDouble
 MISSION 	: Inserts Poets To Double 
 			  Hash Table
 FORMULA 	: { 193.NAME[i] + (i+1).227 } % 233
 COLLISION	: NODE --> NODE +
 		i. [ { 181.NAME[j] + (i+1).233 } % 199 ]
***********************************************/
int InsertDouble(Table *table, const char name[], const int key){
	int nod = 0, collision = 0, w;
	int strlength = strlen(name);
	for(int m = 0; m < strlength; m++)
		nod += ((191*abs((int)(name[m])) + (m+1)*197)%227)%MAX_HASH;
	nod %= MAX_HASH;
	w = 0;
	while((*table).node[nod].used){
		w++;
		for(int m = 0; m < strlength; m++)
			nod += (w*((181*abs((int)(name[m])) + (w+1)*233)%199))%MAX_HASH;
		collision++;
		nod %= MAX_HASH;
	}
	(*table).node[nod].key = key;
	(*table).node[nod].used = true;
	strcpy((*table).node[nod].name, name);
	return collision + 1;
}

/***********************************************
                SEARCH FUNCTIONS              */
/***********************************************
 FUNCTION 	: SearchLinear
 MISSION 	: Searches Poets From 
 			  Linear Hash Table
***********************************************/
int SearchLinear(Table *table, const char name[], int *index){
	int nod = 0, collision = 1;
	int strlength = strlen(name);
	for(int m = 0; m < strlength; m++)
		nod += ((193*abs((int)(name[m])) + (m+1)*227)%233)%MAX_HASH;
	nod %= MAX_HASH;
	while((*table).node[nod].used){
		if(strcmp((*table).node[nod].name, name) == 0){
			*index = (*table).node[nod].key;
			return collision;
		}
		nod = (nod + 1) % MAX_HASH;
		collision++;
	}
	return -collision;
}

/***********************************************
 FUNCTION 	: SearchQuadratic
 MISSION 	: Searches Poets From 
 			  Quadratic Hash Table
***********************************************/
int SearchQuadratic(Table *table, const char name[], int *index){
	int nod = 0, collision = 1, w = 0;
	int strlength = strlen(name);
	for(int m = 0; m < strlength; m++)
		nod += ((181*abs((int)(name[m])) + (m+1)*233)%199)%MAX_HASH;
	nod %= MAX_HASH;
	while((*table).node[nod].used){
		if(strcmp((*table).node[nod].name, name) == 0){
			*index = (*table).node[nod].key;
			return collision;
		}
		w++;
		nod = (nod + 199*w + 211*w*w) % MAX_HASH;
		collision++;
	}
	return -collision;
}

/***********************************************
 FUNCTION 	: SearchDouble
 MISSION 	: Searches Poets From 
 			  Double Hash Table
***********************************************/
int SearchDouble(Table *table, const char name[], int *index){
	int nod = 0, collision = 1, w = 0;
	int strlength = strlen(name);
	for(int m = 0; m < strlength; m++)
		nod += ((191*abs((int)(name[m])) + (m+1)*197)%227)%MAX_HASH;
	nod %= MAX_HASH;
	while((*table).node[nod].used){
		if(strcmp((*table).node[nod].name, name) == 0){
			*index = (*table).node[nod].key;
			return collision;
		}
		w++;
		for(int m = 0; m < strlength; m++)
			nod += (w*((181*abs((int)(name[m])) + (w+1)*233)%199))%MAX_HASH;
		collision++;
		nod %= MAX_HASH;
	}
	return -collision;
}

/***********************************************
               PRINT FUNTIONS                 */
/***********************************************
 FUNCTION 	: printPoem
 MISSION 	: Reads The Poems From Database 
 			  and Prints Them To The Screen
***********************************************/
bool printPoem(int n){
	FILE *f;
	int length;
	char c;
	if(f = fopen("database.txt","r")){
		fseek(f, (long)n, SEEK_CUR);
		fscanf(f, "%d", &length);
		for(int i = 0; i < length; i++){
			fscanf(f, "%c", &c);
			cout << c;
			if(c == '\n') i++;
		}
		fclose(f);
		return true;
	}else
		return false;
}

/***********************************************
 FUNCTION 	: menu
 MISSION 	: Prints The Main Menu
***********************************************/
void menu(){
	cout << endl;
	cout << "  ---------------------------------" << endl;
	cout << "  |  Search Linear Hash     -  1  |" << endl;
	cout << "  |  Search Quadratic Hash  -  2  |" << endl;
	cout << "  |  Search Double Hash     -  3  |" << endl;
	cout << "  |  Show All Poets         -  4  |" << endl;
	cout << "  |  Exit                   -  0  |" << endl;
	cout << "  ---------------------------------" << endl << endl << "  >> ";
}


/***********************************************
            MAIN FUNCTION                     */
/***********************************************
 FUNCTION 	: MAIN
 MISSION 	: START OF THE MAIN FUNCTION
***********************************************/
int main(){
	Table 	myTmp, myNoPoem, myLinear, myQuadratic, myDouble;
	int 	collision, rnd, index;
	float 	totalColl[6] = {0.0};
	char 	c;
	bool 	myMenu = true;
	
	if(!ReadData(&myTmp, &myNoPoem)){
		cout << "database.txt Could Not Be Found..." << endl;
		system("PAUSE");
		return -1;
	}
	//PrintTmpData(&myTmp, &myNoPoem);
	
	srand(time(NULL));
	cout << "            LINEAR        QUADRATIC       DOUBLE" << endl;
	cout << "LOAD    succ    unsuc   succ    unsuc   succ    unsuc" << endl;
	cout << "----    -----   -----   -----   -----   -----   -----" << endl;
	for(int i=0; i<MAX_POET; i++){
		collision = InsertLinear(&myLinear, myTmp.node[i].name, myTmp.node[i].key);
		collision = InsertQuadratic(&myQuadratic, myTmp.node[i].name, myTmp.node[i].key);
		collision = InsertDouble(&myDouble, myTmp.node[i].name, myTmp.node[i].key);
		if((i+1)%((MAX_HASH)/19) == 0 && i > 0){
			for(int j=0; j<100; j++){
				rnd = rand()%(i+1);
				collision = SearchLinear(&myLinear, myTmp.node[rnd].name, &index);
				totalColl[0] += collision;
				collision = SearchQuadratic(&myQuadratic, myTmp.node[rnd].name, &index);
				totalColl[2] += collision;
				collision = SearchDouble(&myDouble, myTmp.node[rnd].name, &index);
				totalColl[4] += collision;
				
				rnd = rand()%MNO_POEM;
				collision = SearchLinear(&myLinear, myNoPoem.node[rnd].name, &index);
				totalColl[1] -= collision;
				collision = SearchQuadratic(&myQuadratic, myNoPoem.node[rnd].name, &index);
				totalColl[3] -= collision;
				collision = SearchDouble(&myDouble, myNoPoem.node[rnd].name, &index);
				totalColl[5] -= collision;
			}
			cout << "%" << 5*(i+1)/7 << "\t";
			for(int k = 0; k < 6; k++){
				totalColl[k] /= 100.0;
				printf("%3.3f\t", totalColl[k]);
			}
			cout << endl;
		}
	}
	cout << "---------------------------------------------------------" << endl;
	
	// MENU LOOP
	while(myMenu){
		menu();
		c = getch();
		cout << c << endl;
		rnd = 0;
		switch(c){
			case '0':
				myMenu = false;
				break;
			case '1':
			case '2':
			case '3':
				cout << "Num Of Poet ( to See The Poem ) : ";
				cin >> rnd;
				rnd = (rnd + MAX_POET - 1)%MAX_POET;
				cout << endl;
				if(c == '1')
					collision = SearchLinear(&myLinear, myTmp.node[rnd].name, &index);
				else if(c == '2')
					collision = SearchQuadratic(&myQuadratic, myTmp.node[rnd].name, &index);
				else 
					collision = SearchDouble(&myDouble, myTmp.node[rnd].name, &index);
				cout << "Collision Number : " << collision << endl;
				cout << endl << "----------------------------------------" << endl;
				cout << "Poet : " << myTmp.node[rnd].name << endl;
				cout << "Poem : " << endl;
				cout << "----------------------------------------" << endl;
				if(!printPoem(index)){
					cout << "database.txt Could Not Be Found..." << endl;
					system("PAUSE");
					return -1;
				}
				cout << endl << "----------------------------------------" << endl;
				break;
			default:
				cout << endl << endl;
				PrintTmpData(&myTmp, &myNoPoem);
				cout << endl;
				break;
		}
		cout << endl;
	}
	
	cout << endl;
	system("PAUSE");
	return 0;
}
